/**************************************************************

comfun.h (C-Munipack project)
Common functions, definitions, etc.
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef COMFUN_H
#define COMFUN_H

#include "../config.h"
#include "cmunipack.h"

/* ************* Common constats *******************/

// Standard strings
#define PKGNAME 		"C-Munipack"
#define WARRANTYLINE 	"This program is free but without waranty; see the software license for details.\n"
#define COPYRIGHTLINE 	"Copyright (C) 2005 David Motl, Czech republic\n"
#define PROJECTLINE 	"This program is part of C-Munipack project.\n"
#define HOMEPAGELINE 	"Home page: http://c-munipack.sourceforge.net/\n"
#define BUGREPORTS 		"Bug reports: David Motl <dmotl@volny.cz>\n"

// Max length of the line in text files
#define MAXLINE 1024
#define MAXBUFFER 65536

// Pi number
#define PI 3.14159265358979323846

// Machine dependent constants
#define MACHMAX 1.0E10
        // machine maximum real number (appropriately)
#define MACHEPS 1.0E-15
        // machine epsilon real number (appropriately)

/* *************** Global variables ****************/

extern int cmpack_quiet;
	// Supress messages except error reports
extern int cmpack_debug;
	// Be verbose - for debugging purposes

/* **************** Common macros ******************/

#define fsig(x)   ((x)>=0?(+1):(-1))
#define fabs(x)   ((x)>=0?(x):-(x))
#define fmax(x,y) (((x)>(y))?(x):(y))
#define fmin(x,y) (((x)<(y))?(x):(y))

/**************** Common functions *****************/

// Prints short version of GPL license to the stdout.
void PrintLicense();

// Prints package version and short info
void PrintVersion(const char *program);

// Prints package version and short info
void PrintPrologue(const char *program);

// Prints human readable form of error code to the specified file.
void PrintError(const char *program, const char *file, int aline, int status);

// Default callback function which prints messages to stdout.
void PrintMessage(const char *msg);

// This internal functions should be used for all print outs. It performs message
// filtering and pass it to the callback procedure.
void PrintOut(int debug, const char *message);

// Makes copy of specified file (infile to outfile)
// Returns zero on success, negative value on failure.
int copyfile(const char *infile, const char *outfile);

#ifndef _WIN32
// If the directory doesn't exists, makes one. Returns nonzero on success
// (directory exists or was made). Returns zero otherwise.
int chkdir(const char *path);
#endif

// Converts file mask to format string
void MaskToFormat(char *format, const char *mask);

// Uppercases the string
char *strup(char *s);

// Changes extension of inf file to ext, results stores to outf
char *changeext(char *outf, const char *inf, const char *ext);

// Finds a string needle in array haystack (uses memmem)
const char *memcasestr(const char *haystack, size_t haystacklen, const char *needle);

#endif
