/**************************************************************

matsolve.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/
#define _NO_OLDNAMES

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "comfun.h"
#include "quicksort.h"
#include "cholesky.h"
#include "robmean.h"
#include "match.h"
#include "matfun.h"
#include "matsolve.h"

static void Init1(int i1, int i2)
/* Construct list of stars in image 1 and sorts them with respect of distance from the center of the abscissa
   given by stars i1 and i2. Adds i1 and i2 to the polygon id1. Returns squared distance of i1 and i2 as z1. */
{
    double xc, yc, dx, dy;
    int j;

    xc = 0.5*(mmatch_x1[i1]+mmatch_x1[i2]);
    yc = 0.5*(mmatch_y1[i1]+mmatch_y1[i2]);
    dx = mmatch_x1[i1]-mmatch_x1[i2];
    dy = mmatch_y1[i1]-mmatch_y1[i2];
    mmatch_z1 = (dx*dx)+(dy*dy);
    for (j=0;j<mmatch_n1;j++) {
	    mmatch_idr1[j] = j;
	    if (j==i1) 
	        mmatch_r1[j] = -2.0; 
	    else if (j==i2) 
	        mmatch_r1[j] = -1.0; 
	    else {
		    dx = mmatch_x1[j]-xc;
		    dy = mmatch_y1[j]-yc;
	        mmatch_r1[j] = (dx*dx)+(dy*dy);
        }
    }
    QuickSortD(mmatch_r1,mmatch_n1,(void **)mmatch_idr1);
}

static void Init2(int j1, int j2)
/* Construct list of stars in image 2 and sorts them with respect of distance from the center of the abscissa
   given by stars i1 and i2. Adds i1 and i2 to the polygon id1. Returns squared distance of i1 and i2 as z1. */
{
    double xc, yc, zoom, dx, dy;
    int j;

    xc = 0.5*(mmatch_x2[j1]+mmatch_x2[j2]);
    yc = 0.5*(mmatch_y2[j1]+mmatch_y2[j2]);
    dx = mmatch_x2[j1]-mmatch_x2[j2];
    dy = mmatch_y2[j1]-mmatch_y2[j2];
    mmatch_z2 = (dx*dx)+(dy*dy);
    zoom = mmatch_z1/mmatch_z2;
    for (j=0;j<mmatch_n2;j++) {
	    mmatch_idr2[j] = j;
	    if (j==j1) 
	        mmatch_r2[j] = -2.0; 
	    else if (j==j2) 
	        mmatch_r2[j] = -1.0; 
	    else {
		    dx = mmatch_x2[j]-xc;
		    dy = mmatch_y2[j]-yc;
	        mmatch_r2[j] = zoom*((dx*dx)+(dy*dy));
        }
    }
    QuickSortD(mmatch_r2,mmatch_n2,(void **)mmatch_idr2);
}

static int Resrov(int n, double c, double mirror, double axi, double ayi, double a2, double *b, double *m)
/* Resi soustavu rovnic a vypocte vyslednou transformacni matici M */
{
	double a[5*4];

	A(0,0) = n*c;
	A(1,0) = 0.0;
	A(2,0) = c*axi;
	A(3,0) = c*ayi;
	A(0,1) = 0.0;
	A(1,1) = A(0,0);
	A(2,1) = mirror*c*ayi;
	A(3,1) = -mirror*c*axi;
	A(0,2) = A(2,0);
	A(1,2) = A(2,1);
	A(2,2) = c*a2;
	A(3,2) = 0.0;
	A(0,3) = A(3,0);
	A(1,3) = A(3,1);
	A(2,3) = 0.0;
	A(3,3) = A(2,2);
	A(4,0) = c*b[0];
	A(4,1) = c*b[1];
	A(4,2) = c*b[2];
	A(4,3) = c*b[3];
	if (Cholesky(4,a)) 
	    return 1;
	if (fabs(A(4,2)*A(4,2)+A(4,3)*A(4,3)-1.0)>0.2) 
	    return 1;

	M(0,0) = A(4,2);
    M(0,1) = mirror*A(4,3);
    M(0,2) = A(4,0);
    M(1,0) = -mirror*A(4,3);
    M(1,1) = A(4,2);
    M(1,2) = A(4,1);
    M(2,0) = 0.0;
    M(2,1) = 0.0;
    M(2,2) = 1.0;
	return 0;
}

static int FindTrafo(int ns, double zoom, double *m, int *mstar, double *sumsq, double *tol)
/* Find coefficients of transformation */
{
      int    i,j,j0,nn,id1,id10,id11,id2,id12,id20,id21,id22;
      double m1,m2,mirror,a0,a1,b0,b1,s,axi,ayi,a2,xx1,yy1,xx2,yy2,xx,yy,tol2,dr,rmin,p[4];

      // We need at least three stars
      if (ns<3)
          return 0;				

      // Mirroring 
      id10 = mmatch_id1[0];  
      id11 = mmatch_id1[1];  
      id12 = mmatch_id1[2];  
      a0 = mmatch_x1[id11]-mmatch_x1[id10];
      a1 = mmatch_y1[id11]-mmatch_y1[id10];
      b0 = mmatch_x1[id12]-mmatch_x1[id10];
      b1 = mmatch_y1[id12]-mmatch_y1[id10];
      m1   = a0*b1-a1*b0;
      id20 = mmatch_id2[0];
      id21 = mmatch_id2[1];
      id22 = mmatch_id2[2];
      a0 = mmatch_x2[id21]-mmatch_x2[id20];
      a1 = mmatch_y2[id21]-mmatch_y2[id20];
      b0 = mmatch_x2[id22]-mmatch_x2[id20];
      b1 = mmatch_y2[id22]-mmatch_y2[id20];
      m2   = a0*b1-a1*b0;
      mirror = fsig(m1*m2);

      // Zero-iteration 
      p[0] = p[1] = p[2] = p[3] = axi = ayi = a2 = 0.0;
      for (i=0; i<ns; i++) {
	      id1  = mmatch_id1[i];
	      xx1  = mmatch_x1[id1];
	      yy1  = mmatch_y1[id1];
	      id2  = mmatch_id2[i];
	      xx2  = mmatch_x2[id2];
	      yy2  = mmatch_y2[id2];
	      axi += xx1;
	      ayi += yy1;
	      a2  += xx1*xx1+yy1*yy1;
	      p[0]+= xx2;
	      p[1]+= yy2;
	      p[2]+= xx1*xx2+mirror*yy1*yy2;
	      p[3]+= yy1*xx2-mirror*xx1*yy2;
      }
      
      // Find estimation of transformation matrix
      if (Resrov(ns,zoom,mirror,axi,ayi,a2,p,m)!=0)
	      return 0;

      // Residuals 
      s = 0.0;
      for (i=0;i<ns;i++) {
	      xx1 = mmatch_x1[mmatch_id1[i]];
	      yy1 = mmatch_y1[mmatch_id1[i]];
	      Trafo(m,xx1,yy1,&xx,&yy);
	      xx2 = mmatch_x2[mmatch_id2[i]];
	      yy2 = mmatch_y2[mmatch_id2[i]];
	      s  += (xx-xx2)*(xx-xx2)+(yy-yy2)*(yy-yy2);
      }
      tol2 = fmax(0.0001, 3.0*mmatch_clip*mmatch_clip*s/(ns-2.0)+MACHEPS);

      // Reziduals for all stars 
      nn = 0; j0 = -1;
      p[0] = p[1] = p[2] = p[3] = axi = ayi = a2 = s = 0.0;
      for (i=0;i<mmatch_n1;i++) {
	      Trafo(m,mmatch_x1[i],mmatch_y1[i],&xx,&yy);
          // Find star with the smallest residues 
          rmin = MACHMAX;
          for (j=0;j<mmatch_n2;j++) {
	          dr =(xx-mmatch_x2[j])*(xx-mmatch_x2[j])+(yy-mmatch_y2[j])*(yy-mmatch_y2[j]);
	          if (dr<rmin) { rmin = dr; j0 = j; }
          }
          j = j0;
          if (rmin<=tol2) {
	          // Star was found 
	          s   += rmin;
    	      xx1  = mmatch_x1[i];
	          yy1  = mmatch_y1[i];
	          xx2  = mmatch_x2[j];
	          yy2  = mmatch_y2[j];
	          axi += xx1;
	          ayi += yy1;
	          a2  += xx1*xx1+yy1*yy1;
	          p[0]+= xx2;
	          p[1]+= yy2;
	          p[2]+= xx1*xx2+mirror*yy1*yy2;
	          p[3]+= yy1*xx2-mirror*xx1*yy2;
	          nn++;
          }
      }
      
      // We need at least three stars
      if (nn<3)
	      return 0;
	      
	  // Find improved transformation matrix
      if (Resrov(nn,zoom,mirror,axi,ayi,a2,p,m)!=0)
	      return 0;
	      
	  // Compute improved tolerance
      tol2 = fmax(0.0001, 3.0*mmatch_clip*mmatch_clip*s/(nn-2.0)+MACHEPS);

      // Test the transformation, comute final tolerance and number of identified stars
      nn=0; s=0.0;
      for (i=0;i<mmatch_n1;i++) {
	      Trafo(m,mmatch_x1[i],mmatch_y1[i],&xx,&yy);
          // Find star with the smallest residues
          rmin = MACHMAX;
          for (j=0;j<mmatch_n2;j++) {
	          dr =(xx-mmatch_x2[j])*(xx-mmatch_x2[j])+(yy-mmatch_y2[j])*(yy-mmatch_y2[j]);
	          if (dr<rmin) { rmin = dr; }
          }
          if (rmin<tol2) {
	          s += rmin;
	          nn++;
          }
	}
	
	// Transformation is OK, if there are at least three stars identified
	if (nn<3) 
		return 0;
      
	// Normal return, transformation is OK
	*mstar = nn;          
	*sumsq = s;
	*tol = fmax(0.0001, 3.0*mmatch_clip*mmatch_clip*s/(nn-2.0)+MACHEPS);
	return 1;
}

static int TestMatch(int n, double *zoom)
/* Mame uz pozadovany pocet hvezd v polygonech id1 a id2 -> testuj polygon */
{
	int    i1, i2, i3, i, j, k, l, rep;
	double s, maxdev, sig, dx1, dx2, dy1, dy2;
	double a1[2], a2[2], b1[2], b2[2];
	double xmed, xsig, xdev, xarc, prum;

	// We need at least three stars
	if (n<3)
		return 0;
	
	/* Calculation of ratio of similarity (scale). */
	l = 0; rep = 0;
	for (i=0;i<n;i++) {
		for (j=i+1;j<n;j++) {
			i1  = mmatch_id1[i];
			i2  = mmatch_id1[j];
			dx1 = mmatch_x1[i2]-mmatch_x1[i1];
			dy1 = mmatch_y1[i2]-mmatch_y1[i1];
            i1  = mmatch_id2[i];
            i2  = mmatch_id2[j];
            dx2 = mmatch_x2[i2]-mmatch_x2[i1];
            dy2 = mmatch_y2[i2]-mmatch_y2[i1];
            mmatch_k2[l] = sqrt((dx2*dx2+dy2*dy2)/(dx1*dx1+dy1*dy1));
            if ((l==0)&&(mmatch_k2[l]<0.9)) rep = 1;
            if (rep) mmatch_k2[l] = 1.0/mmatch_k2[l];
			l++;
		}
	}

	/* Calculation of mean and median */
	cmpack_robustmeanD(l,mmatch_k2,&xmed,&xsig);

    /* Calculation of absolute deviations */
    for (i=0;i<l;i++) {
	    mmatch_dev[i] = fabs(mmatch_k2[i]-xmed);
	    if (mmatch_dev[i] > 0.2*xmed) 
	    	return 0;   // deviation is too great (because it isn't from the true series)
    }

	/* Calculation of the median deviation */
	cmpack_robustmeanD(l,mmatch_dev,&xdev,&xsig);

	if (xdev > 0.05*xmed) 
		return 0;       // Mean deviation is too great

    /* Calculation of the aritmetical mean */
    s = 0.0;
    for (i=0;i<l;i++) 
    	s += mmatch_k2[i];
    prum = s/l;
    /* Find the greatest deviation into mean */
    j = 0; maxdev = -1.0;
    for (i=0;i<l;i++) {
	    mmatch_dev[i] = (mmatch_k2[i]-prum)*(mmatch_k2[i]-prum);
        if (mmatch_dev[i]>maxdev) {
	        maxdev = mmatch_dev[i];
	        j = i;
        }
    }
    /* Calculation of the new mean */
    s = 0.0;
    for (i=0;i<l;i++) 
    	if (i!=j) s += mmatch_k2[i];
    prum = s/(l-1);
    /* Estimation of the standard deviation */
    s = 0.0;
    for (i=0;i<l;i++) {
	    mmatch_dev[i] = (mmatch_k2[i]-prum);
        if (i!=j) s += mmatch_dev[i]*mmatch_dev[i];
    }
    sig = sqrt(s/(l-2))+MACHEPS;
    if (sig > 0.05*prum) 
    	return 0;   /* Standard deviation is too great */
    if (fabs(prum-xmed)>=sig) 
    	return 0; /* Test for equality of median and mean */

    /* Test on the same arcs of N-polygon */
    l = 0;
	for (i=0;i<n;i++) {
		for (j=i+1;j<n;j++) {
		    for (k=j+1;k<n;k++) {
			    i1 = mmatch_id1[i];
			    i2 = mmatch_id1[j];
			    i3 = mmatch_id1[k];
			    a1[0] = mmatch_x1[i2]-mmatch_x1[i1];
			    a1[1] = mmatch_y1[i2]-mmatch_y1[i1];
			    b1[0] = mmatch_x1[i3]-mmatch_x1[i1];
			    b1[1] = mmatch_y1[i3]-mmatch_y1[i1];
			    i1 = mmatch_id2[i];
			    i2 = mmatch_id2[j];
			    i3 = mmatch_id2[k];
			    a2[0] = mmatch_x2[i2]-mmatch_x2[i1];
			    a2[1] = mmatch_y2[i2]-mmatch_y2[i1];
			    b2[0] = mmatch_x2[i3]-mmatch_x2[i1];
			    b2[1] = mmatch_y2[i3]-mmatch_y2[i1];
			    mmatch_k2[l] = Uhel(a1,b1)/Uhel(a2,b2);
			    if ((mmatch_k2[l]>2.0)||(mmatch_k2[l]<0.5)) 
			    	return 0; /* devation onto 1.0 is too great */
			    l++;
		    }
	    }
    }

	/* Calculation of mean and median of arcs */
	cmpack_robustmeanD(l,mmatch_k2,&xarc,&xsig);

    /* Calculation of the deviation into median */
    for (i=0;i<l;i++) {
	    mmatch_dev[i] = fabs(mmatch_k2[i]-xarc);
        if (mmatch_dev[i] > 0.2*xarc) 
        	return 0;
    }
	cmpack_robustmeanD(l,mmatch_dev,&xdev,&xsig);

    if (xdev+MACHEPS > 0.2*xarc) 
    	return 0;           /* Mean deviation is too great */
    if (fabs(xarc-1.0) > (3.0*(xdev+MACHEPS))) 
    	return 0; 			/* Median must be equal 1.0 */

	/* All test are OK */
    if (rep) *zoom = 1.0/xmed;
        else *zoom = xmed;
    return 1;
}

int Solve(int *m_stars)
{
	int 	mstar, nstar, i, j, j0, ni, i1, i2, i3, j1, j2, j3;
	double 	dr, rr, rl, rg, rmax, dif, u1, u2, v1, v2, s;
	double 	xx, yy, dx, dy, sumsq, tol, zoom, m[9];
	char 	msg[256];

	sprintf(msg,"Matching algorithm               : Standard");
	PrintOut(0, msg);
	
	// Clear output
	*m_stars = 0;
    mmatch_offx = 0.0;
    mmatch_offy = 0.0;
    for (i=0; i<mmatch_c2; i++) 
    	mmatch_id[i] = -1;
	
	// Number of stars used for matching
	mmatch_n1 = (mmatch_c1>mmatch_maxstar ? mmatch_maxstar : mmatch_c1);
	mmatch_n2 = (mmatch_c2>mmatch_maxstar ? mmatch_maxstar : mmatch_c2);

	// For all doubles of stars on reference frame -> i1, i2
	for (i1=0;i1<mmatch_n1;i1++) {
		for (i2=i1+1;i2<mmatch_n1;i2++) {
			Init1(i1,i2);
		    mmatch_id1[0] = mmatch_idr1[0];
		    mmatch_id1[1] = mmatch_idr1[1];
			// For all doubles of stars on input frame -> j1, j2
			for (j1=0;j1<mmatch_n2;j1++) {
				for (j2=j1+1;j2<mmatch_n2;j2++) {
					Init2(j1,j2);
				    mmatch_id2[0] = mmatch_idr2[0];
				    mmatch_id2[1] = mmatch_idr2[1];
  				    // For all stars on reference frame except i1 and i2 -> i3
					ni = 2;
					for (i3=2; i3<mmatch_n1 && ni<mmatch_maxstar; i3++) {
                        // Find the star on input frame with the distance most close to i3 -> j3
                        rr   = mmatch_r1[i3];
                        rl   = rr/1.5;
	                	rg   = rr*1.5;
                        rmax = MACHMAX;
                        j3  = -1;
                        for (j=2; j<mmatch_n2 && mmatch_r2[j]<rg; j++) {
                            if (mmatch_r2[j]>rl && mmatch_idr2[j]!=mmatch_id2[ni-1] && mmatch_idr2[j]!=mmatch_id2[ni-2]) {
                            	dif = fabs(mmatch_r2[j]-rr);
	                    		if (dif<rmax) { 
		                    		rmax = dif; 
		                    		j3 = j; 
		                    	}
	                    	}
                        }
                        if (j3<0) {
	                        // Not found, try next i3
                        	continue; 
                    	}
                        // Add i3 and j3 to the polygon 
	                    mmatch_id1[ni] = mmatch_idr1[i3];
                        mmatch_id2[ni] = mmatch_idr2[j3];
                        // Calculation of the triangle coordinates (u1,v1) in the reference file
                        UV(mmatch_id1[ni-2],mmatch_id1[ni-1],mmatch_id1[ni],mmatch_x1,mmatch_y1,&u1,&v1);
                        // Calculation of the triangle coordinates (u2,v2) in the input file 
                        UV(mmatch_id2[ni-2],mmatch_id2[ni-1],mmatch_id2[ni],mmatch_x2,mmatch_y2,&u2,&v2);
                        if (((u1-u2)*(u1-u2)+(v1-v2)*(v1-v2))<0.0005) {
                            // Transformation is found
	                        ni++;
                            if (ni>=mmatch_nstar) {
	                            // We have enough vertices in the polygon
	                            if (TestMatch(ni,&zoom)) {
									// Find transformation matrix 
								    if (FindTrafo(ni,zoom,m,&mstar,&s,&tol)) {
									    // Add transformation to the stack
								        StAdd(ni,s,tol,mstar,m);
									}
								}
	                            break;	// Next j2
                            }
					    }
                    }
                }
            }
        }
    }

	// Select from stack the best case 
    StSelect(&mstar,&sumsq,&tol,m);
    if (mstar==0) {
	    // No coincidences found
	    PrintOut(0, "Coincidences not found!");
	    return MAT_MATCH_NOT_FOUND;
    }
    
    // Matching OK
	sprintf(msg,"Sum of square in the best case   : %.4f", sumsq);
	PrintOut(0, msg);
	sprintf(msg,"Tolerance                        : %.2f", sqrt(tol));
	PrintOut(0, msg);
	sprintf(msg,"Transformation matrix            : ");
	PrintOut(0, msg);
	sprintf(msg,"   %15.3f %15.3f %15.3f",M(0,0),M(0,1),M(0,2));
	PrintOut(0, msg);
	sprintf(msg,"   %15.3f %15.3f %15.3f",M(1,0),M(1,1),M(1,2));
	PrintOut(0, msg);
	sprintf(msg,"   %15.3f %15.3f %15.3f",M(2,0),M(2,1),M(2,2));
	PrintOut(0, msg);
	
	// Set cross-references for all stars on input frame
	nstar = 0;
    for (i=0; i<mmatch_c1; i++) {
	    // Compute estimated destination coordinates
        Trafo(m,mmatch_x1[i],mmatch_y1[i],&xx,&yy);
        // Select the nearest star 
        rr = MACHMAX; j0 = -1;
        for (j=0; j<mmatch_c2; j++) {
	        if (mmatch_id[j]<0) {
		        dx = xx-mmatch_x2[j];
		        dy = yy-mmatch_y2[j];
		        dr = (dx*dx+dy*dy);
		        if (dr<rr) { rr = dr; j0 = j; }
	        }
        }
        if ((j0>=0)&&(rr<tol)) {
	        mmatch_id[j0] = i;
	        nstar++;
        }
    }
    *m_stars = nstar;
    mmatch_offx = M(0,2);
    mmatch_offy = M(1,2);

    // Normal return	
	return 0;
}
